//
// Metallic Flakes Shader
// Copyright (c) Microsoft Corporation. All rights reserved.
//
// Note: This effect file works with EffectEdit.
//
// NOTE:
// The metallic surface consists of 2 layers - 
// 1. a polished layer of wax on top (contributes a smooth specular reflection and an environment mapped reflection with a Fresnel term), and 
// 2. a blue metallic layer with a sprinkling of gold metallic flakes underneath


// sparkle parameters
#define SPRINKLE    1.0
#define SCATTER     0.5

#define VOLUME_NOISE_SCALE  10

// scene info (EffectEdit specific controls)
string XFile = "teapot.x";
int    BCLR  = 0xff404080;

half time : TIME;

// textures
texture EnvironmentMap 
< 
    string type = "CUBE"; 
    string name = "lobbycube.dds"; 
>;

// procedural texture that contains a normal map used for the metal sparkles
texture NoiseMap 
< 
    string type = "VOLUME"; 
    string function = "GenerateSparkle"; 
    int width = 32;
    int height = 32;
    int depth = 32; 
>;

texture NoiseMap2D
< 
//    string type = "TEXTURE"; 
//    string function = "GenerateSparkle2D";
    string name = "noise.bmp";
    int width = 32;
    int height = 32;
//    int depth = 32; 
>;

// transformations
half4x3 WorldView  : WORLDVIEW;
half4x4 Projection : PROJECTION;

// light direction (view space)
half3 L < string UIDirectional = "Light Direction"; > = normalize(half3(-0.397f, -0.397f, 0.827f));

// light intensity
half4 I_a = { 0.3f, 0.3f, 0.3f, 1.0f };    // ambient
half4 I_d = { 1.0f, 0.0f, 0.0f, 1.0f };    // diffuse
half4 I_s = { 0.7f, 0.7f, 0.7f, 1.0f };    // specular

// material reflectivity
half4 k_a : MATERIALAMBIENT = { 0.2f, 0.2f, 0.2f, 1.0f };    // ambient  (metal)
half4 k_d : MATERIALDIFFUSE = { 0.1f, 0.1f, 0.9f, 1.0f };    // diffuse  (metal)
half4 k_s = { 1.0f, 0.9f, 0.1f, 1.0f };    // specular (metal)
half4 k_r = { 0.7f, 0.7f, 0.7f, 1.0f };    // specular (wax)

// function used to fill the volume noise texture
half4 GenerateSparkle(half3 Pos : POSITION) : COLOR
{
    half4 Noise = (half4)0;

    // set the normal with a probability based on SPRINKLE
    if(SPRINKLE >= abs(noise(Pos * 600)))
    {
        // scatter the normal (in vertex space) based on SCATTER
        Noise.rgb = half3(1 - SCATTER * abs(noise(Pos * 500)), SCATTER * noise((Pos + 1) * 500), SCATTER * noise((Pos + 2) * 500));
        Noise.rgb = normalize(Noise.rgb);
    }

    // bias the normal
    Noise.rgb = (Noise.rgb + 1)/2;

    // diffuse noise
    Noise.w = abs(noise(Pos * 15)) * 0.75 + 0.25;
///    Noise.w = abs(noise(Pos * 5)) * 0.75 + 0.25; // weird lava-ish texture effect

    return Noise;
}


struct VS_OUTPUT
{
    half4 Position   : POSITION;
    half3 Diffuse    : COLOR0;
    half3 Specular   : COLOR1;               
    half3 Reflection : TEXCOORD0;               
    half3 NoiseCoord : TEXCOORD1;               
    half3 Glossiness : TEXCOORD2;               
    half3 HalfVector : TEXCOORD3;
};

// vertex shader
VS_OUTPUT VS(
    half3 Position : POSITION,
    half3 Normal   : NORMAL, 
    half3 Tangent  : TANGENT)
{
    VS_OUTPUT Out = (VS_OUTPUT)0;

    L = -L;

    half3 P = mul(half4(Position, 1), (half4x3)WorldView);   // position (view space)
    half3 N = normalize(mul(Normal, (half3x3)WorldView));     // normal (view space)
    half3 T = normalize(mul(Tangent, (half3x3)WorldView));    // tangent (view space)
    half3 B = cross(N, T);                                     // binormal (view space)
    half3 R = normalize(2 * dot(N, L) * N - L);                // reflection vector (view space)
    half3 V = -normalize(P);                                   // view direction (view space)
    half3 G = normalize(2 * dot(N, V) * N - V);                // glance vector (view space)
    half3 H = normalize(L + V);                                // half vector (view space)
    half  f = 0.5 - dot(V, N); f = 1 - 4 * f * f;              // fresnel term

    // position (projected)
    Out.Position = mul(half4(P, 1), Projection);             

    // diffuse + ambient (metal)
    Out.Diffuse = I_a * k_a + I_d * k_d * max(0, dot(N, L)); 

    // specular (wax)
    Out.Specular  = saturate(dot(H, N));
    Out.Specular = pow(Out.Specular,32);
    Out.Specular *= k_r;                                       

     // glossiness (wax)
    Out.Glossiness = f * k_r;

    // transform half vector into vertex space
    Out.HalfVector = half3(dot(H, N), dot(H, B), dot(H, T));   
    Out.HalfVector = (1 + Out.HalfVector) / 2;  // bias

    // environment cube map coordinates
    Out.Reflection = half3(-G.x, G.y, -G.z);                   

    // volume noise coordinates
    Out.NoiseCoord = Position * VOLUME_NOISE_SCALE;             

    return Out;
}


struct VSforFF_OUTPUT
{
    half4 Position   : POSITION;
    half4 Diffuse    : COLOR0; // diffuse + specular
    half2 TexCoord   : TEXCOORD; // noise
    half3 TexCoord2  : TEXCOORD1; // envmap
};

// vertex shader whose output goes to the fixed function pipeline
VSforFF_OUTPUT VSforFF(
    half3 Position : POSITION,
    half3 Normal   : NORMAL,
    half3 Tangent : TANGENT)
{
    VSforFF_OUTPUT Out = (VSforFF_OUTPUT)0;

    L = -L;

    half3 P = mul(half4(Position, 1), (half4x3)WorldView);   // position (view space)
    half3 N = normalize(mul(Normal, (half3x3)WorldView));     // normal (view space)
    half3 R = normalize(2 * dot(N, L) * N - L);                // reflection vector (view space)
    half3 V = -normalize(P);                                   // view direction (view space)
    half3 G = normalize(2 * dot(N, V) * N - V);                // glance vector (view space)
    half3 H = normalize(L + V);                                // half vector (view space)
    half  f = 0.5 - dot(V, N); f = 1 - 4 * f * f;              // fresnel term

    // position (projected)
    Out.Position = mul(half4(P, 1), Projection);             

    // specular (wax)
    half Specular  = saturate(dot(H, N));
    Specular = pow(Specular,32);
    Specular *= k_r;

    // transform half vector into vertex space
    half  HalfVector = dot(H, N);   
    HalfVector = (1 + HalfVector) / 2;  // bias

    // environment cube map coordinates
    Out.TexCoord2 = half3(-G.x, G.y, -G.z);

    // noise texture coordinates
//    Out.TexCoord = half2((Normal.x + Normal.y), (Normal.y + Normal.z))*VOLUME_NOISE_SCALE;
//    Out.TexCoord = half2(int((Position.x*44)),0)*8;

    Out.TexCoord = half2(Position.x,Position.y)*8
                 + half2(cos(Position.y),Position.z)*-7
                 + half2(sin(Position.z),cos(Position.x));
//Out.TexCoord += frac(cos(Position.xy)*50)*1;

//    Out.TexCoord = normalize(Tangent*Position);

    Out.Diffuse = max(0, dot(N, L));

    half4 Sparkle = saturate(dot((saturate(HalfVector) - 0.5) * 2, (.99 - 0.5) * 2));
    Sparkle = pow(Sparkle, 16) * k_s;
    Sparkle -= Sparkle * I_d*Out.Diffuse; // reduce oversaturation in the areas between sparkle and diffuse
//    Sparkle -= Sparkle * I_d+k_s*1.1; // weird
//    Sparkle = clamp(Sparkle,0,0);


    // diffuse + ambient (metal)
    Out.Diffuse = I_a * k_a + I_d * k_d * Out.Diffuse ;
    Out.Diffuse += Specular + Sparkle;

//    Out.Diffuse.rgb = ((half)pow(Out.Diffuse.rgb,0.25) + (half)pow(Out.Diffuse.rgb,4))/2;

    Out.Diffuse.r = (half)((half)pow(Out.Diffuse.r,0.25) + (half)pow(Out.Diffuse.r,4))/2;
    Out.Diffuse.g = (half)((half)pow(Out.Diffuse.g,0.25) + (half)pow(Out.Diffuse.g,4))/2;
    Out.Diffuse.b = (half)((half)pow(Out.Diffuse.b,0.25) + (half)pow(Out.Diffuse.b,4))/2;

     // glossiness (wax)
    Out.Diffuse.a = f * k_r;


    return Out;
}

// samplers
sampler Environment = sampler_state
{ 
    Texture = (EnvironmentMap);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

sampler SparkleNoise = sampler_state
{ 
    Texture = (NoiseMap);
    MipFilter = LINEAR; 
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

sampler SparkleNoise2D = sampler_state
{ 
    Texture = (NoiseMap2D);
    MipFilter = LINEAR; 
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

// pixel shader
half4 PS(VS_OUTPUT In) : COLOR
{   
    half4 Color = (half4)0;
    half3 Diffuse, Specular, Gloss, Sparkle;

    // volume noise
    half4 Noise = tex3D(SparkleNoise, In.NoiseCoord);

    // noisy diffuse of metal
    Diffuse = In.Diffuse * Noise.a;
    
    // glossy specular of wax
    Specular = pow(In.Specular, 2);
    
    // glossy reflection of wax 
    Gloss = texCUBE(Environment, In.Reflection) * saturate(In.Glossiness);              

    // specular sparkle of flakes
    Sparkle = saturate(dot((saturate(In.HalfVector) - 0.5) * 2, (Noise.rgb  - 0.5) * 2));
    Sparkle = pow(Sparkle, 16) * k_s;

    // combine the contributions
    Color.rgb = Diffuse + Specular + Gloss + Sparkle;
    Color.w   = 1;

    return Color;
}  

// technique
technique TMetallicFlakes
{
    pass P0
    {
        VertexShader = compile vs_1_1 VS();
        PixelShader  = compile ps_1_1 PS();

        AlphaBlendEnable = FALSE;
        CullMode         = CCW;
    }
}

// technique
technique TMetallicFlakesNoPS
{
    pass P0
    {
        VertexShader = compile vs_1_1 VSforFF();
        PixelShader = NULL;

        DitherEnable = TRUE;
        AlphaBlendEnable = FALSE;
        CullMode         = CCW;


        AlphaBlendEnable = FALSE;
        CullMode         = CCW;
        DitherEnable = TRUE;

        // texture
        Texture[0] = (NoiseMap2D);
        Texture[1] = (EnvironmentMap);

        // sampler states
        MinFilter[0] = LINEAR;
        MagFilter[0] = LINEAR;
        MipFilter[0] = NONE;

        MinFilter[1] = LINEAR;
        MagFilter[1] = LINEAR;
        MipFilter[1] = LINEAR;

        // set up texture stage states for single texture modulated by diffuse 
        ColorOp[0]   = ADDSIGNED2X;
        ColorArg1[0] = TEXTURE;
        ColorArg2[0] = DIFFUSE;
        AlphaOp[0]   = DISABLE;

        ColorOp[1]   = BLENDDIFFUSEALPHA;
        ColorArg1[1] = TEXTURE;
        ColorArg2[1] = CURRENT;
        AlphaOp[1]   = DISABLE;

       AlphaBlendEnable = FALSE;
        SrcBlend         = ONE;
        DestBlend        = ZERO;
//        BlendOp = SUBTRACT;

    }
}

